lexer grammar v3;
@members {
    /** The log to be used for error diagnostics.
     */
    private Log log;
    
    static final byte NO_INSERT_SEMI = 0; // default
    static final byte INSERT_SEMI = 1; 
    static final byte IGNORE_FOR_SEMI = 2; 
    static final byte[] semiKind = new byte[LAST_TOKEN];
    { 
      for (int i = SEMI_INSERT_START; i < SEMI_INSERT_END; ++i) {
          semiKind[i] = INSERT_SEMI;
      }
      semiKind[RBRACE] = INSERT_SEMI;
      semiKind[STRING_LITERAL] = INSERT_SEMI;
      semiKind[QUOTE_LBRACE_STRING_LITERAL] = INSERT_SEMI;
      semiKind[DECIMAL_LITERAL] = INSERT_SEMI;
      semiKind[OCTAL_LITERAL] = INSERT_SEMI;
      semiKind[HEX_LITERAL] = INSERT_SEMI;
      semiKind[TIME_LITERAL] = INSERT_SEMI;
      semiKind[FLOATING_POINT_LITERAL] = INSERT_SEMI;
      semiKind[IDENTIFIER] = INSERT_SEMI;
      
      semiKind[WS] = IGNORE_FOR_SEMI;
      semiKind[COMMENT] = IGNORE_FOR_SEMI;
      semiKind[LINE_COMMENT] = IGNORE_FOR_SEMI;
    }
      
    int previousTokenType = -1;
    final Token syntheticSemi = new CommonToken(SEMI);
    List tokens = new ArrayList();
    
    public v3Lexer(Context context, CharStream input) {
    	this(input);
        this.log = Log.instance(context);
    }
    
    /** Allow emitting more than one token from a lexer rule
     */
    public void emit(Token token) {
        int ttype = token.getType();
        //System.err.println("::: " + ttype + " --- " + token.getText());
        if (previousTokenType == RBRACE && (ttype == EOF || semiKind[ttype] == INSERT_SEMI)) {
            state.token = syntheticSemi;
            tokens.add(syntheticSemi);
            //System.err.println("INSERTING in front of: " + ttype);
        } else {
            state.token = token;
        }
    	tokens.add(token);

        if (ttype != EOF && semiKind[ttype] != IGNORE_FOR_SEMI) {
            previousTokenType = ttype;
        }
    }
    public Token nextToken() {
        if ( tokens.size() > 0 ) {
            return (Token)tokens.remove(0);
        }
    	super.nextToken();
        if ( tokens.size()==0 ) {
            emit(Token.EOF_TOKEN);
        }
        Token cur = (Token)tokens.remove(0);
        return cur;
    }    
    
    /** Track "He{"l{"l"}o"} world" quotes
     */
    private static class BraceQuoteTracker {
        private static BraceQuoteTracker quoteStack = null;
        private int braceDepth;
        private char quote;
        private boolean percentIsFormat;
        private BraceQuoteTracker next;
        private BraceQuoteTracker(BraceQuoteTracker prev, char quote, boolean percentIsFormat) {
            this.quote = quote;
            this.percentIsFormat = percentIsFormat;
            this.braceDepth = 1;
            this.next = prev;
        }
        static void enterBrace(int quote, boolean percentIsFormat) {
            if (quote == 0) {  // exisiting string expression or non string expression
                if (quoteStack != null) {
                    ++quoteStack.braceDepth;
                    quoteStack.percentIsFormat = percentIsFormat;
                }
            } else {
                quoteStack = new BraceQuoteTracker(quoteStack, (char)quote, percentIsFormat); // push
            }
        }
        /** Return quote kind if we are reentering a quote
         * */
        static char leaveBrace() {
            if (quoteStack != null && --quoteStack.braceDepth == 0) {
                return quoteStack.quote;
            }
            return 0;
        }
        static boolean rightBraceLikeQuote(int quote) {
            return quoteStack != null && quoteStack.braceDepth == 1 && (quote == 0 || quoteStack.quote == (char)quote);
        }
        static void leaveQuote() {
            assert (quoteStack != null && quoteStack.braceDepth == 0);
            quoteStack = quoteStack.next; // pop
        }
        static boolean percentIsFormat() {
            return quoteStack != null && quoteStack.percentIsFormat;
        }
        static void resetPercentIsFormat() {
            quoteStack.percentIsFormat = false;
        }
        static boolean inBraceQuote() {
            return quoteStack != null;
        }
    }
    
    void processString() {
    	setText(StringLiteralProcessor.convert( log, getCharIndex(), getText() ));
    }

    void processTranslationKey() {
        String text = getText().substring(2); // remove '##'
        if (text.length() > 0) {
            text = StringLiteralProcessor.convert( log, getCharIndex(), text );
        }
        setText(text);
    }

    // quote context --
    static final int CUR_QUOTE_CTX	= 0;	// 0 = use current quote context
    static final int SNG_QUOTE_CTX	= 1;	// 1 = single quote quote context
    static final int DBL_QUOTE_CTX	= 2;	// 2 = double quote quote context
 }
@header {
package com.sun.tools.javafx.antlr;

import com.sun.tools.javac.util.Context;
import com.sun.tools.javac.util.Log;
}

ABSTRACT : 'abstract' ;
ASSERT : 'assert' ;
ATTRIBUTE : 'attribute' ;
BIND : 'bind' ;
BREAK : 'break' ;
CLASS : 'class' ;
CONTINUE : 'continue' ;
DELETE : 'delete' ;
FALSE : 'false' ;
FOR : 'for' ;
FUNCTION : 'function' ;
IF : 'if' ;
IMPORT : 'import' ;
INIT : 'init' ;
INSERT : 'insert' ;
LET : 'let' ;
NEW : 'new' ;
NOT : 'not' ;
NULL : 'null' ;
PACKAGE : 'package' ;
POSTINIT : 'postinit' ;
PRIVATE : 'private' ;
PROTECTED : 'protected' ;
PUBLIC : 'public' ;
READONLY : 'readonly' ;
RETURN : 'return' ;
SUPER : 'super' ;
SIZEOF : 'sizeof' ;
STATIC : 'static' ;
THIS : 'this' ;
THROW : 'throw' ;
TRY : 'try' ;
TRUE : 'true' ;
VAR : 'var' ;
WHILE : 'while' ;
POUND : '#' ;
LPAREN : '(' ;
LBRACKET : '[' ;
PLUSPLUS : '++' ;
SUBSUB : '--' ;
PIPE : '|' ;
AFTER : 'after' ;
AND : 'and' ;
AS : 'as' ;
BEFORE : 'before' ;
CATCH : 'catch' ;
ELSE : 'else' ;
EXCLUSIVE : 'exclusive' ;
EXTENDS : 'extends' ;
FINALLY : 'finally' ;
FIRST : 'first' ;
FROM : 'from' ;
IN : 'in' ;
INDEXOF : 'indexof' ;
INSTANCEOF : 'instanceof' ;
INTO : 'into' ;
INVERSE : 'inverse' ;
LAST : 'last' ;
LAZY : 'lazy' ;
ON : 'on' ;
OR : 'or' ;
REPLACE : 'replace' ;
REVERSE : 'reverse' ;
STEP : 'step' ;
THEN : 'then' ;
TYPEOF : 'typeof' ;
WITH : 'with' ;
WHERE : 'where' ;
DOTDOT : '..' ;
RPAREN : ')' ;
RBRACKET : ']' ;
SEMI : ';' ;
COMMA : ',' ;
DOT : '.' ;
EQEQ : '==' ;
EQ : '=' ;
GT : '>' ;
LT : '<' ;
LTGT : '<>' ;
LTEQ : '<=' ;
GTEQ : '>=' ;
PLUS : '+' ;
SUB : '-' ;
STAR : '*' ;
SLASH : '/' ;
PERCENT : '%' ;
PLUSEQ : '+=' ;
SUBEQ : '-=' ;
STAREQ : '*=' ;
SLASHEQ : '/=' ;
PERCENTEQ : '%=' ;
COLON : ':' ;
QUES : '?' ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 344
STRING_LITERAL  		: '"' DoubleQuoteBody '"'  	{ processString(); }
				| '\'' SingleQuoteBody '\''  	{ processString(); }
				;
// String Expression token implementation
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 348
QUOTE_LBRACE_STRING_LITERAL 	: '"' DoubleQuoteBody '{'   	{ processString(); }
				  NextIsPercent[DBL_QUOTE_CTX] 
				| '\'' SingleQuoteBody '{'   	{ processString(); }
				  NextIsPercent[SNG_QUOTE_CTX] 
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 353
LBRACE				: '{'				{ BraceQuoteTracker.enterBrace(0, false); } 
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 355
RBRACE_QUOTE_STRING_LITERAL 	:				{ BraceQuoteTracker.rightBraceLikeQuote(DBL_QUOTE_CTX) }?=>
				  '}' DoubleQuoteBody '"'	{ BraceQuoteTracker.leaveBrace(); 
				         			  BraceQuoteTracker.leaveQuote(); 
				         			  processString(); }
				|				{ BraceQuoteTracker.rightBraceLikeQuote(SNG_QUOTE_CTX) }?=>
				  '}' SingleQuoteBody '\''	{ BraceQuoteTracker.leaveBrace(); 
				         			  BraceQuoteTracker.leaveQuote(); 
				         			  processString(); }
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 364
RBRACE_LBRACE_STRING_LITERAL 	:				{ BraceQuoteTracker.rightBraceLikeQuote(DBL_QUOTE_CTX) }?=>
				  '}' DoubleQuoteBody '{'	{ BraceQuoteTracker.leaveBrace(); 
				         			  processString(); }
				   NextIsPercent[CUR_QUOTE_CTX]	
				|				{ BraceQuoteTracker.rightBraceLikeQuote(SNG_QUOTE_CTX) }?=>
				  '}' SingleQuoteBody '{'	{ BraceQuoteTracker.leaveBrace(); 
				         			  processString(); }
				   NextIsPercent[CUR_QUOTE_CTX]	
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 373
RBRACE				:				{ !BraceQuoteTracker.rightBraceLikeQuote(CUR_QUOTE_CTX) }?=>
				  '}'				{ BraceQuoteTracker.leaveBrace(); }
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 376
fragment
DoubleQuoteBody  :	 (~('{' |'"'|'\\')|'\\' .)*  ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 379
fragment
SingleQuoteBody  :	 (~('{' |'\''|'\\')|'\\' .)*  ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 382
fragment
NextIsPercent[int quoteContext]
	 			: ((' '|'\r'|'\t'|'\u000C'|'\n')* '%')=>
	 							{ BraceQuoteTracker.enterBrace(quoteContext, true); }
				|				{ BraceQuoteTracker.enterBrace(quoteContext, false); }
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 388
FORMAT_STRING_LITERAL		: 				{ BraceQuoteTracker.percentIsFormat() }?=>
				  '%' (~' ')* 			{ BraceQuoteTracker.resetPercentIsFormat(); }
				;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 391
TRANSLATION_KEY                 : '##' 
                                  ( 
                                    '[' TranslationKeyBody ']'
                                  )?                            { processTranslationKey(); }
				;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 397
fragment
TranslationKeyBody              : (~('[' | ']' | '\\')|'\\' .)+
                                ;
 
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 401
TIME_LITERAL : (DECIMAL_LITERAL | Digits '.' (Digits)? (Exponent)? ) ( 'ms' | 'm' | 's' | 'h' ) ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 403
DECIMAL_LITERAL : ('0' | '1'..'9' '0'..'9'*) ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 405
OCTAL_LITERAL : '0' ('0'..'7')+ ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 407
HEX_LITERAL : '0' ('x'|'X') HexDigit+    			{ setText(getText().substring(2, getText().length())); };

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 409
fragment
HexDigit : ('0'..'9'|'a'..'f'|'A'..'F') ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 412
FLOATING_POINT_LITERAL
    :     d=DECIMAL_LITERAL RangeDots
    	  	{
    	  		$d.setType(DECIMAL_LITERAL);
    	  		emit($d);
          		$RangeDots.setType(DOTDOT);
    	  		emit($RangeDots);
    	  	}
    |     d=OCTAL_LITERAL RangeDots
    	  	{
    	  		$d.setType(OCTAL_LITERAL);
    	  		emit($d);
          		$RangeDots.setType(DOTDOT);
    	  		emit($RangeDots);
    	  	}
    |	  Digits '.' (Digits)? (Exponent)? 
    | '.' Digits (Exponent)? 
    |     Digits Exponent
    ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 432
fragment
RangeDots 
	:	DOTDOT
	;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 436
fragment
Digits	:	('0'..'9')+ 
 	;
// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 439
fragment
Exponent : 	('e'|'E') ('+'|'-')? Digits
 	;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 443
IDENTIFIER 
	: Letter (Letter|JavaIDDigit)*
	| '<<' (~'>'| '>' ~'>')* '>'* '>>'			{ setText(getText().substring(2, getText().length()-2)); }
	;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 448
fragment
Letter
    :  '\u0024' |
       '\u0041'..'\u005a' |
       '\u005f' |
       '\u0061'..'\u007a' |
       '\u00c0'..'\u00d6' |
       '\u00d8'..'\u00f6' |
       '\u00f8'..'\u00ff' |
       '\u0100'..'\u1fff' |
       '\u3040'..'\u318f' |
       '\u3300'..'\u337f' |
       '\u3400'..'\u3d2d' |
       '\u4e00'..'\u9fff' |
       '\uf900'..'\ufaff'
    ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 465
fragment
JavaIDDigit
    :  '\u0030'..'\u0039' |
       '\u0660'..'\u0669' |
       '\u06f0'..'\u06f9' |
       '\u0966'..'\u096f' |
       '\u09e6'..'\u09ef' |
       '\u0a66'..'\u0a6f' |
       '\u0ae6'..'\u0aef' |
       '\u0b66'..'\u0b6f' |
       '\u0be7'..'\u0bef' |
       '\u0c66'..'\u0c6f' |
       '\u0ce6'..'\u0cef' |
       '\u0d66'..'\u0d6f' |
       '\u0e50'..'\u0e59' |
       '\u0ed0'..'\u0ed9' |
       '\u1040'..'\u1049'
   ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 484
WS  :  (' '|'\r'|'\t'|'\u000C'|'\n') {$channel=HIDDEN;}
    ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 487
COMMENT
    :   '/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 491
LINE_COMMENT
    : '//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;}
    ;

// $ANTLR src "src/share/classes/com/sun/tools/javafx/antlr/v3.g" 495
LAST_TOKEN
    :	'~~~~~~~~' {false}? '~~~~~~~~'
    ;
    	
/*------------------------------------------------------------------
 * PARSER RULES
 *------------------------------------------------------------------*/

